@@ -1,10 +1,11 @@
 """The cimsparql.model module contains the base class CimModel."""
 
+import asyncio
 import re
 from dataclasses import dataclass
 from functools import cached_property
 from string import Template
-from typing import Dict, Optional
+from typing import Dict, List, Optional
 
 import pandas as pd
 
@@ -34,7 +35,7 @@ from cimsparql.data_models import (
     TransformerWindingSchema,
     WindGeneratingUnitsSchema,
 )
-from cimsparql.graphdb import GraphDBClient, ServiceConfig
+from cimsparql.graphdb import AsyncGraphDBClient, GraphDBClient, ServiceConfig
 from cimsparql.type_mapper import TypeMapper
 
 
@@ -53,7 +54,7 @@ class Model:
     ) -> None:
         self.client = client
         self.config = config or ModelConfig()
-        self.mapper = mapper or TypeMapper(client)
+        self.mapper = mapper or TypeMapper(client.service_cfg)
 
     @staticmethod
     def _col_map(data_row) -> Dict[str, str]:
@@ -74,11 +75,13 @@ class Model:
         pref = self.client.prefixes
         return "cim" in pref and self.mapper.have_cim_version(pref["cim"])
 
-    def get_table_and_convert(
-        self, query: str, index: Optional[str] = None, columns: Optional[Dict[str, str]] = None
+    def _convert_result(
+        self,
+        result: pd.DataFrame,
+        data_row: List[dict],
+        index: Optional[str] = None,
+        columns: Optional[Dict[str, str]] = None,
     ) -> pd.DataFrame:
-        result, data_row = self.client.get_table(query)
-
         col_map = self.col_map(data_row, columns)
         result = self.mapper.map_data_types(result, col_map)
 
@@ -86,6 +89,15 @@ class Model:
             return result.set_index(index)
         return result
 
+    async def get_table_and_convert(
+        self, query: str, index: Optional[str] = None, columns: Optional[Dict[str, str]] = None
+    ) -> pd.DataFrame:
+        if isinstance(self.client, AsyncGraphDBClient):
+            result, data_row = await self.client.get_table(query)
+        else:
+            result, data_row = self.client.get_table(query)
+        return self._convert_result(result, data_row, index, columns)
+
     @property
     def state_repo(self) -> str:
         return self.config.system_state_repo or self.client.service_cfg.url
@@ -113,8 +125,7 @@ class CimModel(Model):
     def full_model_query(self) -> str:
         return self.template_to_query(templates.FULL_MODEL_QUERY)
 
-    @cached_property
-    def full_model(self) -> FullModelSchema:
+    async def full_model(self) -> FullModelSchema:
         """Return all models where all depencies has been created and is available
 
         All profiles EQ/SSH/TP/SV will define a md:FullModel with possible dependencies. One profile
@@ -124,7 +135,7 @@ class CimModel(Model):
         Example:
         >> model.full_model()
         """
-        df = self.get_table_and_convert(self.full_model_query)
+        df = await self.get_table_and_convert(self.full_model_query)
         return FullModelSchema(df)
 
     @property
@@ -132,9 +143,8 @@ class CimModel(Model):
         """Market activation date for this repository."""
         return self.template_to_query(templates.MARKET_DATES_QUERY)
 
-    @cached_property
-    def market_dates(self) -> MarketDatesSchema:
-        df = self.get_table_and_convert(self.market_dates_query, index="mrid")
+    async def market_dates(self) -> MarketDatesSchema:
+        df = await self.get_table_and_convert(self.market_dates_query, index="mrid")
         return MarketDatesSchema(df)
 
     def bus_data_query(self, region: Optional[str] = None) -> str:
@@ -145,39 +155,40 @@ class CimModel(Model):
         substitutes = {"region": region or ".*"}
         return self.template_to_query(templates.THREE_WINDING_DUMMY_NODES_QUERY, substitutes)
 
-    def bus_data(self, region: Optional[str] = None) -> BusDataSchema:
+    async def bus_data(self, region: Optional[str] = None) -> BusDataSchema:
         """Query name of topological nodes (TP query).
 
         Args:
            region: Limit to region (use None to get all)
         """
-        df = pd.concat(
-            [
-                self.get_table_and_convert(query(region), index="node")
-                for query in [self.bus_data_query, self.three_winding_dummy_nodes_query]
-            ]
+        dfs = await asyncio.gather(
+            self.get_table_and_convert(self.bus_data_query(region), index="node"),
+            self.get_table_and_convert(self.three_winding_dummy_nodes_query(region), index="node"),
         )
+        df = pd.concat(dfs)
         return BusDataSchema(df)
 
     def loads_query(self, region: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*", "ssh_graph": self.config.ssh_graph}
         return self.template_to_query(templates.LOADS_QUERY, substitutes)
 
-    def loads(self, region: Optional[str] = None) -> LoadsSchema:
+    async def loads(self, region: Optional[str] = None) -> LoadsSchema:
         """Query load data.
 
         Args:
            region: regexp that limits to region
         """
         query = self.loads_query(region)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return LoadsSchema(df)
 
     def wind_generating_units_query(self, region: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*"}
         return self.template_to_query(templates.WIND_GENERATING_UNITS_QUERY, substitutes)
 
-    def wind_generating_units(self, region: Optional[str] = None) -> WindGeneratingUnitsSchema:
+    async def wind_generating_units(
+        self, region: Optional[str] = None
+    ) -> WindGeneratingUnitsSchema:
         """Query wind generating units.
 
         Args:
@@ -191,23 +202,23 @@ class CimModel(Model):
 
         """
         query = self.wind_generating_units_query(region)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return WindGeneratingUnitsSchema(df)
 
     def synchronous_machines_query(self, region: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*", "ssh_graph": self.config.ssh_graph}
         return self.template_to_query(templates.SYNCHRONOUS_MACHINES_QUERY, substitutes)
 
-    def synchronous_machines(self, region: Optional[str] = None) -> SynchronousMachinesSchema:
+    async def synchronous_machines(self, region: Optional[str] = None) -> SynchronousMachinesSchema:
         query = self.synchronous_machines_query(region)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return SynchronousMachinesSchema(df)
 
     def connections_query(self, region: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*"}
         return self.template_to_query(templates.CONNECTIONS_QUERY, substitutes)
 
-    def connections(self, region: Optional[str] = None) -> ConnectionsSchema:
+    async def connections(self, region: Optional[str] = None) -> ConnectionsSchema:
         """Query connectors
 
         Args:
@@ -224,14 +235,14 @@ class CimModel(Model):
             >>> model.connections()
         """
         query = self.connections_query(region)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return ConnectionsSchema(df)
 
     def borders_query(self, region: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*"}
         return self.template_to_query(templates.BORDERS_QUERY, substitutes)
 
-    def borders(self, region: Optional[str] = None) -> BordersSchema:
+    async def borders(self, region: Optional[str] = None) -> BordersSchema:
         """Retrieve ACLineSegments where one terminal is inside and the other is outside the region
 
         Args:
@@ -239,14 +250,14 @@ class CimModel(Model):
             limit: return first 'limit' number of rows
         """
         query = self.borders_query(region)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return BordersSchema(df)
 
     def exchange_query(self, region: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*"}
         return self.template_to_query(templates.EXCHANGE_QUERY, substitutes)
 
-    def exchange(self, region: Optional[str] = None) -> ExchangeSchema:
+    async def exchange(self, region: Optional[str] = None) -> ExchangeSchema:
         """Retrieve ACLineSegments where one terminal is inside and the other is outside the region.
 
         Args:
@@ -259,16 +270,16 @@ class CimModel(Model):
             df = pd.DataFrame([], columns=cols, index=index)
             return ExchangeSchema(df)
         query = self.exchange_query(region)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return ExchangeSchema(df)
 
     def converters_query(self, region: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*", "ssh_graph": self.config.ssh_graph}
         return self.template_to_query(templates.CONVERTERS_QUERY, substitutes)
 
-    def converters(self, region: Optional[str] = None) -> ConvertersSchema:
+    async def converters(self, region: Optional[str] = None) -> ConvertersSchema:
         query = self.converters_query(region)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return ConvertersSchema(df)
 
     def transformers_connected_to_converter_query(self, region: Optional[str] = None) -> str:
@@ -277,7 +288,7 @@ class CimModel(Model):
             templates.TRANSFORMERS_CONNECTED_TO_CONVERTER_QUERY, substitutes
         )
 
-    def transformers_connected_to_converter(
+    async def transformers_connected_to_converter(
         self, region: Optional[str] = None
     ) -> TransfConToConverterSchema:
         """Query list of transformer connected at a converter (Voltage source or DC)
@@ -287,14 +298,16 @@ class CimModel(Model):
 
         """
         query = self.transformers_connected_to_converter_query(region)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return TransfConToConverterSchema(df)
 
     def ac_lines_query(self, region: Optional[str] = None, rate: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
         return self.template_to_query(templates.AC_LINE_QUERY, substitutes)
 
-    def ac_lines(self, region: Optional[str] = None, rate: Optional[str] = None) -> AcLinesSchema:
+    async def ac_lines(
+        self, region: Optional[str] = None, rate: Optional[str] = None
+    ) -> AcLinesSchema:
         """Query ac line segments
 
         Args:
@@ -307,7 +320,7 @@ class CimModel(Model):
             >>> model.ac_lines()
         """
         query = self.ac_lines_query(region, rate)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return AcLinesSchema(df)
 
     def series_compensators_query(
@@ -316,7 +329,7 @@ class CimModel(Model):
         substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
         return self.template_to_query(templates.SERIES_COMPENSATORS_QUERY, substitutes)
 
-    def series_compensators(
+    async def series_compensators(
         self, region: Optional[str] = None, rate: Optional[str] = None
     ) -> BranchComponentSchema:
         """Query series compensators
@@ -325,14 +338,14 @@ class CimModel(Model):
            region: Limit to region
         """
         query = self.series_compensators_query(region, rate)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return BranchComponentSchema(df)
 
     def transformers_query(self, region: Optional[str] = None, rate: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
         return self.template_to_query(templates.TRANSFORMERS_QUERY, substitutes)
 
-    def transformers(
+    async def transformers(
         self, region: Optional[str] = None, rate: Optional[str] = None
     ) -> TransformersSchema:
         """Query transformer windings.
@@ -341,7 +354,7 @@ class CimModel(Model):
            region: Limit to region
         """
         query = self.transformers_query(region, rate)
-        df = self.get_table_and_convert(query)
+        df = await self.get_table_and_convert(query)
         return TransformersSchema(df)
 
     def two_winding_transformers_query(
@@ -354,7 +367,7 @@ class CimModel(Model):
         substitutes = {"region": region or ".*"}
         return self.template_to_query(templates.TWO_WINDING_ANGLE_QUERY, substitutes)
 
-    def two_winding_transformers(
+    async def two_winding_transformers(
         self, region: Optional[str] = None, rate: Optional[str] = None
     ) -> TransformerWindingSchema:
         """Query two-winding transformer.
@@ -370,8 +383,8 @@ class CimModel(Model):
         """
         query = self.two_winding_transformers_query(region, rate)
         query_angle = self.two_winding_angle_query(region)
-        data = self.get_table_and_convert(query, index="mrid")
-        angle = self.get_table_and_convert(query_angle, index="mrid")
+        data = await self.get_table_and_convert(query, index="mrid")
+        angle = await self.get_table_and_convert(query_angle, index="mrid")
         if not angle.empty:
             data["angle"] += angle.reindex(index=data.index, fill_value=0.0).squeeze()
         return TransformerWindingSchema(data)
@@ -386,7 +399,7 @@ class CimModel(Model):
         substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
         return self.template_to_query(templates.THREE_WINDING_QUERY, substitutes)
 
-    def three_winding_transformers(
+    async def three_winding_transformers(
         self, region: Optional[str] = None, rate: Optional[str] = None
     ) -> TransformerWindingSchema:
         """Query three-winding transformer. Return as three two-winding transformers.
@@ -398,9 +411,9 @@ class CimModel(Model):
             >>> model.two_winding_transformers()
         """
         query = self.three_winding_transformers_query(region, rate)
-        data = self.get_table_and_convert(query, index="mrid")
+        data = await self.get_table_and_convert(query, index="mrid")
         query_loss = self.three_winding_loss_query(region)
-        loss = self.get_table_and_convert(query_loss, index="mrid")
+        loss = await self.get_table_and_convert(query_loss, index="mrid")
         df = pd.concat([data.assign(ploss_1=0.0), loss.loc[data.index]], axis=1)
         return TransformerWindingSchema(df)
 
@@ -408,66 +421,61 @@ class CimModel(Model):
     def substation_voltage_level_query(self) -> str:
         return self.template_to_query(templates.SUBSTATION_VOLTAGE_LEVEL_QUERY)
 
-    @cached_property
-    def substation_voltage_level(self) -> SubstationVoltageSchema:
+    async def substation_voltage_level(self) -> SubstationVoltageSchema:
         query = self.substation_voltage_level_query
-        df = self.get_table_and_convert(query, index="substation")
+        df = await self.get_table_and_convert(query, index="substation")
         return SubstationVoltageSchema(df)
 
     @property
     def disconnected_query(self) -> str:
         return self.template_to_query(templates.DISCONNECTED_QUERY)
 
-    @cached_property
-    def disconnected(self) -> DisconnectedSchema:
+    async def disconnected(self) -> DisconnectedSchema:
         """Query disconneced status from ssh profile (not available in GraphDB)."""
-        df = self.get_table_and_convert(self.disconnected_query)
+        df = await self.get_table_and_convert(self.disconnected_query)
         return DisconnectedSchema(df)
 
     @property
     def powerflow_query(self) -> str:
         return self.template_to_query(templates.POWER_FLOW_QUERY)
 
-    @cached_property
-    def powerflow(self) -> PowerFlowSchema:
+    async def powerflow(self) -> PowerFlowSchema:
         """Query powerflow from sv profile (not available in GraphDB)."""
-        df = self.get_table_and_convert(self.powerflow_query, index="mrid")
+        df = await self.get_table_and_convert(self.powerflow_query, index="mrid")
         return PowerFlowSchema(df)
 
     @property
     def coordinates_query(self) -> str:
         return self.template_to_query(templates.COORDINATES_QUERY)
 
-    @property
-    def coordinates(self) -> CoordinatesSchema:
-        return CoordinatesSchema(self.get_table_and_convert(self.coordinates_query))
+    async def coordinates(self) -> CoordinatesSchema:
+        return CoordinatesSchema(await self.get_table_and_convert(self.coordinates_query))
 
     @property
     def st_group_codes_names_query(self) -> str:
         return self.template_to_query(templates.STATION_GROUP_CODE_NAME_QUERY)
 
-    @property
-    def station_group_codes_and_names(self) -> StationGroupCodeNameSchema:
-        df = self.get_table_and_convert(self.st_group_codes_names_query)
+    async def station_group_codes_and_names(self) -> StationGroupCodeNameSchema:
+        df = await self.get_table_and_convert(self.st_group_codes_names_query)
         return StationGroupCodeNameSchema(df.set_index("station_group"))
 
     def branch_node_withdraw_query(self, region: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*"}
         return self.template_to_query(templates.BRANCH_NODE_WITHDRAW_QUERY, substitutes)
 
-    def branch_node_withdraw(self, region: Optional[str] = None) -> BranchWithdrawSchema:
+    async def branch_node_withdraw(self, region: Optional[str] = None) -> BranchWithdrawSchema:
         """Query branch flow from sv profile."""
         query = self.branch_node_withdraw_query(region)
-        df = self.get_table_and_convert(query, index="mrid")
+        df = await self.get_table_and_convert(query, index="mrid")
         return BranchWithdrawSchema(df)
 
     def dc_active_flow_query(self, region: Optional[str] = None) -> str:
         substitutes = {"region": region or ".*"}
         return self.template_to_query(templates.DC_ACTIVE_POWER_FLOW_QUERY, substitutes)
 
-    def dc_active_flow(self, region: Optional[str] = None) -> DcActiveFlowSchema:
+    async def dc_active_flow(self, region: Optional[str] = None) -> DcActiveFlowSchema:
         query = self.dc_active_flow_query(region)
-        df = self.get_table_and_convert(query)
+        df = await self.get_table_and_convert(query)
         # Unable to group on max within the sparql query so we do it here.
         df = df.iloc[df.groupby("mrid")["p"].idxmax()].set_index("mrid")
         df["p"] *= df["direction"]
@@ -477,8 +485,7 @@ class CimModel(Model):
     def regions_query(self) -> str:
         return self.template_to_query(templates.REGIONS_QUERY)
 
-    @cached_property
-    def regions(self) -> RegionsSchema:
+    async def regions(self) -> RegionsSchema:
         """Query regions
 
         Property:
@@ -492,7 +499,7 @@ class CimModel(Model):
         """
         # TODO: Probably deprecate this property in the future. But keep for now.
         # We need to find a solution for custom namespaces in queries
-        df = self.get_table_and_convert(self.regions_query, index="mrid")
+        df = await self.get_table_and_convert(self.regions_query, index="mrid")
         return RegionsSchema(df)
 
     def add_mrid_query(self, rdf_type: Optional[str] = None, graph: Optional[str] = None) -> str:
@@ -513,8 +520,18 @@ class CimModel(Model):
 
 
 def get_cim_model(
-    service_cfg: Optional[ServiceConfig] = None, model_cfg: Optional[ModelConfig] = None
+    service_cfg: Optional[ServiceConfig] = None,
+    model_cfg: Optional[ModelConfig] = None,
+    async_sparql_wrapper: bool = False,
 ) -> CimModel:
-    """Get a CIM Model."""
-    client = GraphDBClient(service_cfg)
-    return CimModel(client, model_cfg)
+    """Get a CIM Model.
+
+
+    Args:
+        service_cfg: Configurations for the triple store service
+        model_cfg: Conifugrations for the CIM mode
+        async_sparql_wrapper: If True http calls are made via asynchronous requests.
+            If False, the native SparqlWrapper sends requests via urllib
+    """
+    c = AsyncGraphDBClient(service_cfg) if async_sparql_wrapper else GraphDBClient(service_cfg)
+    return CimModel(c, model_cfg)
